from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from krita import *
import struct
import math

KI = Krita.instance()

XYZ_GRID_SIZE = 32

class XYZGridCell():
    def __init__(self):
        self.tris = list()

class View3DViewport(QOpenGLWidget):
    _gl = None
    _liveCheckbox = None
    _tex = 0
    _texw = 0
    _texh = 0
    _dl = 0
    _verts = []
    _tgtx = 0
    _tgty = 0
    _tgtz = 0
    _rotx = 0
    _roty = 0
    _camd = 10
    _camRot = False
    _camPan = False
    _camZoom = False
    _viewFlat = False
    _viewFilter = False
    _viewWireframe = False
    _viewBackfaces = True
    _viewOrthographic = False
    _viewBgColor = QColor(26, 51, 76)
    _nextUpdateIsSnapshot = 0 # non-zero value is how much to divide
    _doc = None
    _mx = 0
    _my = 0
    _updateTPData = False
    _tiImage = None
    _tiImageOwnerDoc = None
    _tiImageDoc = None
    _fiLastLayer = None
    _periodicUpdates = True

    def __init__(self, parent=None):
        super(View3DViewport, self).__init__(parent)
        fmt = QSurfaceFormat()
        fmt.setRedBufferSize(8)
        fmt.setGreenBufferSize(8)
        fmt.setBlueBufferSize(8)
        fmt.setAlphaBufferSize(8)
        fmt.setSamples(0)
        fmt.setSwapBehavior(QSurfaceFormat.DoubleBuffer)
        fmt.setSwapInterval(0)
        fmt.setMajorVersion(2)
        fmt.setMinorVersion(1)
        fmt.setProfile(QSurfaceFormat.CompatibilityProfile)
        self.setFormat(fmt)
        self._timer = QTimer(self)
        self._timer.timeout.connect(self.timerTick)
        self._timer.start(33)
        self._toggleFilteringAction = QAction("Filtering")
        self._toggleFilteringAction.triggered.connect(self.toggleFiltering)
        self._toggleFilteringAction.setCheckable(True)
        self._toggleFilteringAction.setChecked(self._viewFilter)
        self._toggleShadingAction = QAction("Shading")
        self._toggleShadingAction.triggered.connect(self.toggleShading)
        self._toggleShadingAction.setCheckable(True)
        self._toggleShadingAction.setChecked(not self._viewFlat)
        self._toggleWireframeAction = QAction("Wireframe Overlay")
        self._toggleWireframeAction.triggered.connect(self.toggleWireframe)
        self._toggleWireframeAction.setCheckable(True)
        self._toggleWireframeAction.setChecked(self._viewWireframe)
        self._toggleBackfacesAction = QAction("Backfaces")
        self._toggleBackfacesAction.triggered.connect(self.toggleBackfaces)
        self._toggleBackfacesAction.setCheckable(True)
        self._toggleBackfacesAction.setChecked(self._viewBackfaces)
        self._toggleOrthographicAction = QAction("Orthographic")
        self._toggleOrthographicAction.triggered.connect(self.toggleOrthographic)
        self._toggleOrthographicAction.setCheckable(True)
        self._toggleOrthographicAction.setChecked(self._viewOrthographic)
        self._resetViewAction = QAction("Reset View")
        self._resetViewAction.triggered.connect(self.resetView)
        self._alignCanvasToViewAction = QAction("Align Canvas to View")
        self._alignCanvasToViewAction.triggered.connect(self.alignCanvasToView)
        self._backgroundColorAction = QAction("Background Color...")
        self._backgroundColorAction.triggered.connect(self.backgroundColor)
        self._toImageAction = QAction("Create Image from View")
        self._toImageAction.triggered.connect(self.toImage)
        self._fromImageAction = QAction("Backproject Last Created Image")
        self._fromImageAction.triggered.connect(self.fromImage)
        self._createMaskAction = QAction("Create Mask from View")
        self._createMaskAction.triggered.connect(self.createMask)
        self._createTexCoordMaskAction = QAction("Create Texture Coordinate Mask")
        self._createTexCoordMaskAction.triggered.connect(self.createTexCoordMask)
        self._createTexCoordWireframeAction = QAction("Create Texture Coordinate Wireframe")
        self._createTexCoordWireframeAction.triggered.connect(self.createTexCoordWireframe)
        self._triplanarPasteAction = QAction("Triplanar Paste")
        self._triplanarPasteAction.triggered.connect(self.triplanarPaste)
        self._selectionAsAlphaMaskAction = QAction("Selection as Alpha Mask")
        self._selectionAsAlphaMaskAction.setCheckable(True)
        self._selectionAsAlphaMaskAction.setChecked(True)
        self._setTextureToCurrentDocumentAction = QAction("Set Texture to Current Document")
        self._setTextureToCurrentDocumentAction.triggered.connect(self.setTextureToCurrentDocument)
        self._setTextureToAnyDocumentAction = QAction("Set Texture to Any Document")
        self._setTextureToAnyDocumentAction.triggered.connect(self.setTextureToAnyDocument)
        self._snapshotTo50PercentAction = QAction("Snapshot to 50%% size")
        self._snapshotTo50PercentAction.triggered.connect(self.snapshotTo50Percent)
        self._snapshotTo25PercentAction = QAction("Snapshot to 25%% size")
        self._snapshotTo25PercentAction.triggered.connect(self.snapshotTo25Percent)
        self._snapshotTo125PercentAction = QAction("Snapshot to 12.5%% size")
        self._snapshotTo125PercentAction.triggered.connect(self.snapshotTo125Percent)
        self._openMeshAction = QAction("Open Mesh...")
        self._openMeshAction.triggered.connect(self.openMesh)
        self._openRecentAction = [None]*10
        for i in range(10):
            self._openRecentAction[i] = QAction(KI.readSetting("view3dplugin", "recent" + str(i), ""))
            self._openRecentAction[i].triggered.connect(self.openRecent)
        self._clearRecentAction = QAction("Clear Recent Meshes")
        self._clearRecentAction.triggered.connect(self.clearRecent)

    def msgBox(self, msg):
        box = QMessageBox(self)
        box.setText(msg)
        box.setWindowTitle('View3D Message')
        box.setStandardButtons(QMessageBox.Ok)
        box.exec()

    def triSTBarycentric(self, a, b, c, px, py, vwu):
        v1x = b[6] - a[6]
        v1y = b[7] - a[7]
        v2x = c[6] - a[6]
        v2y = c[7] - a[7]
        v3x = px - a[6]
        v3y = py - a[7]
        d00 = v1x*v1x + v1y*v1y
        d01 = v1x*v2x + v1y*v2y
        d11 = v2x*v2x + v2y*v2y
        d20 = v3x*v1x + v3y*v1y
        d21 = v3x*v2x + v3y*v2y
        den = d00*d11 - d01*d01
        vwu[1] = (d11*d20 - d01*d21)/den
        vwu[2] = (d00*d21 - d01*d20)/den
        vwu[0] = 1.0 - vwu[1] - vwu[2]

    def readJTF(self, fn):
        if self._dl != 0:
            self._gl.glDeleteLists(1, self._dl)
            self._dl = 0
        with open(fn, 'rb') as f:
            # Read header
            header = struct.unpack('<ccccLL', f.read(12))
            if header[0] != b'J' or header[1] != b'T' or header[2] != b'F' or header[3] != b'!':
                self.msgBox('Not a JTF file')
                return
            if header[4] != 0:
                self.msgBox('Unsupported JTF version')
            faces = header[5]
            # Read triangles
            self._verts = [None]*faces*3
            for i in range(faces*3):
                self._verts[i] = struct.unpack('<ffffffff', f.read(32))

    def getDoc(self):
        if self._doc != None:
            return self._doc
        return KI.activeDocument()

    def applySelectionAsAlphaMask(self, pixels):
        if not self._selectionAsAlphaMaskAction.isChecked():
            return
        doc = self.getDoc()
        if doc == None:
            return
        sel = doc.selection()
        if sel == None:
            return
        w = doc.width()
        h = doc.height()
        sela = bytearray(sel.pixelData(0, 0, w, h))
        for i in range(0, w*h):
            a = sela[i]
            idx = i*4
            if a == 0:
                pixels[idx] = pixels[idx + 1] = pixels[idx + 2] = pixels[idx + 3] = 0
            elif a < 255:
                pixels[idx + 3] = int((pixels[idx + 3]/255.0)*(a/255.0)*255.0)

    def createXYZGrid(self):
        self._xyzGridSize = max(self._texw, self._texh)//XYZ_GRID_SIZE + 1
        self._xyzGrid = [None]*(self._xyzGridSize*self._xyzGridSize)
        for i in range(0, self._xyzGridSize*self._xyzGridSize):
            self._xyzGrid[i] = XYZGridCell()
        for i in range(0, len(self._verts), 3):
            a = self._verts[i]
            b = self._verts[i + 1]
            c = self._verts[i + 2]
            minS = max(0, min(self._xyzGridSize - 1, math.floor(min(a[6], b[6], c[6])*self._texw)//XYZ_GRID_SIZE))
            minT = max(0, min(self._xyzGridSize - 1, math.floor(min(a[7], b[7], c[7])*self._texh)//XYZ_GRID_SIZE))
            maxS = max(0, min(self._xyzGridSize - 1, math.ceil(max(a[6], b[6], c[6])*self._texw)//XYZ_GRID_SIZE))
            maxT = max(0, min(self._xyzGridSize - 1, math.ceil(max(a[7], b[7], c[7])*self._texh)//XYZ_GRID_SIZE))
            for gt in range(minT, maxT + 1):
                for gs in range(minS, maxS + 1):
                    c = self._xyzGrid[gt*self._xyzGridSize + gs]
                    c.tris.append(i)
        for i in range(0, self._xyzGridSize*self._xyzGridSize):
            if len(self._xyzGrid[i].tris) == 0:
                self._xyzGrid[i] = None

    def updateTexture(self):
        gl = self._gl
        if not gl:
            return
        if self._tex == 0:
           self._tex = gl.glGenTextures(1)
        gl.glBindTexture(gl.GL_TEXTURE_2D, self._tex)
        doc = self.getDoc()
        if doc:
            img = doc.projection(0, 0, doc.width(), doc.height())
            w = doc.width()
            h = doc.height()
            if self._nextUpdateIsSnapshot > 0:
                self._liveCheckbox.setCheckState(Qt.Unchecked)
                if self._nextUpdateIsSnapshot > 1:
                    w = max(1, int(w/self._nextUpdateIsSnapshot))
                    h = max(1, int(h/self._nextUpdateIsSnapshot))
                    img = img.scaled(w, h, Qt.KeepAspectRatio, Qt.SmoothTransformation)
                self._nextUpdateIsSnapshot = 0
            gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_RGBA, w, h, 0, gl.GL_BGRA, gl.GL_UNSIGNED_BYTE, img.bits().asarray(w*h*4))
            self._texw = w
            self._texh = h
        gl.glBindTexture(gl.GL_TEXTURE_2D, 0)
        self.repaint()

    def timerTick(self):
        if self._periodicUpdates and self.isVisible():
            self.updateTexture()
        if self._doc != None or self._tiImageOwnerDoc != None:
            docFound = False
            iodFound = False
            for d in KI.documents():
                if d == self._doc:
                    docFound = True
                elif d == self._tiImageOwnerDoc:
                    iodFound = True
                if docFound and iodFound:
                    break
            if not docFound:
                self._doc = None
                self._fiLastLayer = None
            if not iodFound:
                self._tiImage = None
                self._tiImageDoc = None
                self._tiImageOwnerDoc = None

    def texCoordAtXY(self, x, y):
        idx = (self.height() - 1 - y) * self.width() + x
        return (int(round(-0.5 + ((int(self._tp[idx*4]*255.0) + int(self._tp[idx*4 + 1]*255)*256)/65535.0)*self._texw)), \
            int(round(-0.5 + ((int(self._tp[idx*4 + 2]*255.0) + int(self._tp[idx*4 + 3]*255)*256)/65535.0)*self._texh)))

    def plotPoint(self, x, y):
        doc = self.getDoc()
        if not doc:
            return
        x, y = self.texCoordAtXY(x, y)
        if x < 0 or y < 0 or x >= doc.width() or y >= doc.height():
            return
        doc.activeNode().setPixelData(QByteArray(b'\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF'), x - 1, y - 1, 3, 3)
        doc.refreshProjection()

    def mouseMoveEvent(self, event):
        super(View3DViewport, self).mouseMoveEvent(event)
        if self._camRot:
            self._rotx += event.globalY() - self._my
            self._roty += event.globalX() - self._mx
            self._mx = event.globalX()
            self._my = event.globalY()
            self.repaint()
        if self._camPan:
            dx = -0.015*(event.globalX() - self._mx)
            dy = 0.015*(event.globalY() - self._my)
            rx = math.radians(self._rotx)
            ry = math.radians(self._roty)
            self._tgtx += math.cos(ry)*dx
            self._tgty += math.cos(rx)*dy
            self._tgtz += math.sin(ry)*dx - math.sin(rx)*dy
            self._mx = event.globalX()
            self._my = event.globalY()
            self.repaint()
        if self._camZoom:
            dy = 0.05*(event.globalY() - self._my)
            ry = math.radians(self._roty)
            self._camd += dy
            self._my = event.globalY()
            self.repaint()

    def mousePressEvent(self, event):
        super(View3DViewport, self).mousePressEvent(event)
        if self._tiImageDoc != None and self._tiImageDoc == KI.activeDocument():
            self._mx = event.globalX()
            self._my = event.globalY()
            return
        if event.button() == Qt.LeftButton:
            if event.modifiers() == Qt.AltModifier:
                self.plotPoint(event.x(), event.y())
                return
            self._camRot = True
            self._mx = event.globalX()
            self._my = event.globalY()
            return
        if event.button() == Qt.MiddleButton:
            self._camPan = True
            self._mx = event.globalX()
            self._my = event.globalY()
            return
        if event.button() == Qt.RightButton:
            self._camZoom = True
            self._mx = event.globalX()
            self._my = event.globalY()
            return

    def mouseReleaseEvent(self, event):
        super(View3DViewport, self).mouseReleaseEvent(event)
        self._updateTPData = True
        self.repaint()
        if event.button() == Qt.LeftButton:
            self._camRot = False
            self._viewRot = False
            return
        if event.button() == Qt.MiddleButton:
            self._camPan = False
            return
        if event.button() == Qt.RightButton:
            self._camZoom = False
            if self._mx == event.globalX() and self._my == event.globalY():
                self.showPopupMenu()
            return

    def showPopupMenu(self):
        # Note: assumes _mx,_my are set to mouse coordinates
        menu = QMenu(self)
        if self._tiImageDoc != None and self._tiImageDoc == KI.activeDocument():
            menu.addAction(self._fromImageAction)
            menu.addSeparator()
        menu.addAction(self._toggleFilteringAction)
        menu.addAction(self._toggleShadingAction)
        menu.addAction(self._toggleWireframeAction)
        menu.addAction(self._toggleBackfacesAction)
        menu.addAction(self._toggleOrthographicAction)
        menu.addSeparator()
        if self._tiImageDoc == None or self._tiImageDoc != KI.activeDocument():
            menu.addAction(self._resetViewAction)
            menu.addAction(self._alignCanvasToViewAction)
        menu.addAction(self._backgroundColorAction)
        menu.addSeparator()
        images = menu.addMenu("Images")
        if self._tiImageDoc == None or self._tiImageDoc != KI.activeDocument():
            images.addAction(self._toImageAction)
        images.addAction(self._fromImageAction)
        images.addSeparator()
        images.addAction(self._createMaskAction)
        images.addAction(self._createTexCoordMaskAction)
        images.addAction(self._createTexCoordWireframeAction)
        images.addSeparator()
        images.addAction(self._triplanarPasteAction)
        images.addSeparator()
        images.addAction(self._selectionAsAlphaMaskAction)
        snapmenu = menu.addMenu("Snapshot")
        snapmenu.addAction(self._snapshotTo50PercentAction)
        snapmenu.addAction(self._snapshotTo25PercentAction)
        snapmenu.addAction(self._snapshotTo125PercentAction)
        menu.addSeparator()
        texmenu = menu.addMenu("Texture")
        texmenu.addAction(self._setTextureToCurrentDocumentAction)
        texmenu.addAction(self._setTextureToAnyDocumentAction)
        menu.addSeparator()
        openmenu = menu.addMenu("Open")
        openmenu.addAction(self._openMeshAction)
        openmenu.addSeparator()
        anyRecent = False
        for i in range(10):
            if len(self._openRecentAction[i].text()) > 0:
                openmenu.addAction(self._openRecentAction[i])
                anyRecent = True
        if anyRecent:
            openmenu.addSeparator()
            openmenu.addAction(self._clearRecentAction)
        menu.popup(QPoint(self._mx, self._my))

    def resetView(self):
        self._tgtx = 0
        self._tgty = 0
        self._tgtz = 0
        self._rotx = 0
        self._roty = 0
        self._camd = 10
        self._updateTPData = True
        self.repaint()

    def alignCanvasToView(self):
        # Note: assumes _mx,_my are set to mouse coordinates
        mousepos = self.mapFromGlobal(QPoint(self._mx, self._my))
        origin = self.texCoordAtXY(mousepos.x(), mousepos.y())
        found = False
        todownx = 0
        todowny = 0
        lastx = -1000000
        lasty = -1000000
        for i in range(0, 16):
            down = self.texCoordAtXY(mousepos.x(), mousepos.y() + i)
            if ((down[0] != origin[0]) or (down[1] != origin[1])) and (down[0] != lastx) and (down[1] != lasty):
                lastx = down[0]
                lasty = down[1]
                todownlen = math.sqrt((down[0] - origin[0])**2 + (down[1] - origin[1])**2)
                if todownlen < 10:
                    todownx += float(down[0] - origin[0])
                    todowny += float(down[1] - origin[1])
                    found = True
        if found:
            todownlen = math.sqrt(todownx**2 + todowny**2)
            todownx /= todownlen
            todowny /= todownlen
            dot = todowny
            det = todownx
            a = math.atan2(det, dot)*180.0/3.1415;
            win = KI.activeWindow()
            if win != None:
                view = win.activeView()
                if view != None:
                    canvas = view.canvas()
                    if canvas != None:
                        restoreOptions = False
                        qwin = Krita.instance().activeWindow().qwindow()
                        if qwin != None:
                            mdi = qwin.findChild(QMdiArea)
                            if mdi != None:
                                scroll = mdi.currentSubWindow().findChild(QAbstractScrollArea)
                                if scroll != None:
                                    viewport = scroll.viewport()
                                    if viewport != None:
                                        restoreOptions = True
                                        zoom = canvas.zoomLevel()
                                        mirror = canvas.mirror()
                                        wrap = canvas.wrapAroundMode()
                                        canvas.setWrapAroundMode(False)
                                        canvas.setMirror(False)
                                        canvas.resetZoom()
                                        canvas.resetRotation()
                                        scroll.horizontalScrollBar().setValue(origin[0] - viewport.width()//2)
                                        scroll.verticalScrollBar().setValue(origin[1] - viewport.height()//2)
                        canvas.setRotation(a)
                        if restoreOptions:
                            canvas.setZoomLevel(zoom)
                            canvas.setMirror(mirror)
                            canvas.setWrapAroundMode(wrap)

    def backgroundColor(self):
        newcol = QColorDialog.getColor(self._viewBgColor, self, "Background Color", QColorDialog.DontUseNativeDialog)
        if newcol.isValid():
            self._viewBgColor = newcol
            self.repaint()

    def toggleShading(self):
        self._viewFlat = not self._viewFlat
        self.repaint()

    def toggleFiltering(self):
        self._viewFilter = not self._viewFilter
        self.repaint()

    def toggleWireframe(self):
        self._viewWireframe = not self._viewWireframe
        self.repaint()

    def toggleBackfaces(self):
        self._viewBackfaces = not self._viewBackfaces
        self.repaint()

    def toggleOrthographic(self):
        self._viewOrthographic = not self._viewOrthographic
        self.repaint()

    def saveRecents(self):
        for i in range(10):
            QAction(KI.writeSetting("view3dplugin", "recent" + str(i), self._openRecentAction[i].text()))

    def addToRecents(self, path):
        for i in range(10):
            if self._openRecentAction[i].text() == path:
                return
        for i in reversed(range(9)):
            self._openRecentAction[i + 1].setText(self._openRecentAction[i].text())
        self._openRecentAction[0].setText(path)
        self.saveRecents()

    def doOpenMesh(self, path):
        self._updateTPData = True
        self.readJTF(path)
        self.createXYZGrid()
        if KI.activeDocument() != None:
            self._doc = KI.activeDocument()

    def openMesh(self):
        fn = QFileDialog.getOpenFileName(self, 'Open Mesh', '', 'JTF files (*.jtf)')
        if fn and len(fn[0].strip()) > 0:
            fn = fn[0].strip()
            self.doOpenMesh(fn)
            self.addToRecents(fn)

    def openRecent(self):
        sender = self.sender()
        self.doOpenMesh(sender.text())

    def clearRecent(self):
        for i in range(10):
            self._openRecentAction[i].setText("")
        self.saveRecents()

    def toImage(self):
        gl = self._gl
        self.makeCurrent()
        gl.glClearColor(0.0, 0.0, 0.0, 0.0)
        self.paintGL()
        gl.glClearColor(self._viewBgColor.redF(), self._viewBgColor.greenF(), self._viewBgColor.blueF(), 0.0)
        rgbapixels = gl.glReadPixels(0, 0, self.width(), self.height(), gl.GL_BGRA, gl.GL_FLOAT)
        self.doneCurrent()
        data = bytearray()
        for rgba in rgbapixels:
            data.append(int(rgba*255.0))
        doc = KI.createDocument(self.width(), self.height(), 'Model', 'RGBA', 'U8', '', 100.0)
        self._tiImageOwnerDoc = doc
        bgrow = bytearray()
        for x in range(self.width()):
            bgrow.append(int(self._viewBgColor.blue()))
            bgrow.append(int(self._viewBgColor.green()))
            bgrow.append(int(self._viewBgColor.red()))
            bgrow.append(int(255))
        bgrow = QByteArray(bgrow)
        node = doc.rootNode().childNodes()[0]
        for y in range(self.height()):
            node.setPixelData(bgrow, 0, y, self.width(), 1)
        node = doc.createNode('Render', 'paintlayer')
        node.setOpacity(255)
        for y in range(self.height()):
            node.setPixelData(QByteArray(data[y*self.width()*4:]), 0, self.height() - 1 - y, self.width(), 1)
        doc.rootNode().addChildNode(node, None)
        node = doc.createGroupLayer('Layers to backproject')
        doc.rootNode().addChildNode(node, None)
        node.setOpacity(255)
        node.setPassThroughMode(False)
        node2 = doc.createNode('Paint Here', 'paintlayer')
        node2.setOpacity(255)
        node.addChildNode(node2, None)
        self._tiImage = node
        self._tiImageDoc = doc
        self._fiLastLayer = None
        KI.activeWindow().addView(doc)
        doc.refreshProjection()

    def fromImage(self):
        if self._tiImage == None:
            self.msgBox('Use "To Image" to create an image first')
            return
        doc = self.getDoc()
        if doc == None:
            self.msgBox('You need an active document')
            return
        rgba = bytearray(self._tiImage.projectionPixelData(0, 0, self.width(), self.height()))
        newrgba = bytearray(b'\x00'*(doc.width()*doc.height()*4))
        newrgbaf = [0.0]*(doc.width()*doc.height()*4)
        newrgbac = [0]*(doc.width()*doc.height())
        minrange = len(newrgba)
        maxrange = 0
        for y in range(self.height()):
            for x in range(self.width()):
                tc = self.texCoordAtXY(x, y)
                if tc[0] >= 0 and tc[0] < doc.width() and tc[1] >= 0 and tc[1] < doc.height():
                    srci = (y*self.width() + x)*4
                    tgti = (tc[1]*doc.width() + tc[0])*4
                    if rgba[srci + 3] > 0.0:
                        newrgbaf[tgti] += rgba[srci]
                        newrgbaf[tgti + 1] += rgba[srci + 1]
                        newrgbaf[tgti + 2] += rgba[srci + 2]
                        newrgbaf[tgti + 3] += rgba[srci + 3]
                        newrgbac[tc[1]*doc.width() + tc[0]] += 1
                        minrange = min(minrange, tc[1]*doc.width() + tc[0])
                        maxrange = max(maxrange, tc[1]*doc.width() + tc[0])
        for i in range(minrange, maxrange + 1):
            if newrgbac[i] > 0:
                newrgba[i*4] = int(newrgbaf[i*4]/newrgbac[i])
                newrgba[i*4 + 1] = int(newrgbaf[i*4 + 1]/newrgbac[i])
                newrgba[i*4 + 2] = int(newrgbaf[i*4 + 2]/newrgbac[i])
                newrgba[i*4 + 3] = int(newrgbaf[i*4 + 3]/newrgbac[i])
        if self._fiLastLayer == None:
            node = doc.createNode('Painted', 'paintlayer')
            node.setOpacity(255)
            doc.rootNode().addChildNode(node, None)
            self._fiLastLayer = node
        else:
            node = self._fiLastLayer
        self.applySelectionAsAlphaMask(newrgba)
        node.setPixelData(QByteArray(newrgba), 0, 0, doc.width(), doc.height())
        doc.refreshProjection()

    def createMask(self):
        doc = self.getDoc()
        if doc == None:
            self.msgBox('You need an active document')
            return
        newrgba = bytearray(b'\x00'*(doc.width()*doc.height()*4))
        for y in range(self.height()):
            for x in range(self.width()):
                tc = self.texCoordAtXY(x, y)
                if tc[0] >= 0 and tc[0] < doc.width() and tc[1] >= 0 and tc[1] < doc.height():
                    tgti = (tc[1]*doc.width() + tc[0])*4
                    newrgba[tgti] = 255
                    newrgba[tgti + 1] = 255
                    newrgba[tgti + 2] = 255
                    newrgba[tgti + 3] = 255
        node = doc.createNode('View Mask', 'paintlayer')
        node.setOpacity(255)
        self.applySelectionAsAlphaMask(newrgba)
        node.setPixelData(QByteArray(newrgba), 0, 0, doc.width(), doc.height())
        doc.rootNode().addChildNode(node, None)
        doc.refreshProjection()

    def createTexCoordLayer(self, solid):
        doc = self.getDoc()
        if doc == None:
            self.msgBox('You need an active document')
            return
        img = QImage(QSize(doc.width(), doc.height()), QImage.Format_ARGB32)
        img.fill(QColor(0, 0, 0, 0))
        poly = QPolygonF()
        poly.append(QPointF(0, 0))
        poly.append(QPointF(0, 0))
        poly.append(QPointF(0, 0))
        painter = QPainter()
        painter.begin(img)
        painter.setRenderHint(QPainter.Antialiasing, False)
        painter.setPen(Qt.white)
        if solid:
            painter.setBrush(Qt.white)
        for i in range(len(self._verts)//3):
            poly[0] = QPointF(self._verts[i*3][6]*doc.width(), self._verts[i*3][7]*doc.height())
            poly[1] = QPointF(self._verts[i*3 + 1][6]*doc.width(), self._verts[i*3 + 1][7]*doc.height())
            poly[2] = QPointF(self._verts[i*3 + 2][6]*doc.width(), self._verts[i*3 + 2][7]*doc.height())
            painter.drawPolygon(poly)
        painter.end()
        bits = img.bits()
        bits.setsize(img.byteCount())
        node = doc.createNode('Tex Coord Mask', 'paintlayer')
        node.setOpacity(255)
        newrgba = bytearray(bits.asarray())
        self.applySelectionAsAlphaMask(newrgba)
        node.setPixelData(QByteArray(newrgba), 0, 0, doc.width(), doc.height())
        doc.rootNode().addChildNode(node, None)
        doc.refreshProjection()

    def createTexCoordMask(self):
        self.createTexCoordLayer(True)

    def createTexCoordWireframe(self):
        self.createTexCoordLayer(False)

    def triplanarPaste(self):
        doc = self.getDoc()
        if doc == None:
            self.msgBox("Need a texture image open")
            return
        sel = doc.selection()
        if self._xyzGrid == None:
            self.msgBox("Missing XYZ grid data, try reopening the mesh with the texture open")
            return
        img = QGuiApplication.clipboard().image()
        if (img == None) or (img.isNull()):
            self.msgBox("An image must be copied to the clipboard")
            return
        ratio = 0.0
        ratiocnt = 0
        for t in range(0, len(self._verts), 3):
            ta = self._verts[t]
            tb = self._verts[t + 1]
            tc = self._verts[t + 2]
            ab3dx = tb[0] - ta[0]
            ab3dy = tb[1] - ta[1]
            ab3dz = tb[2] - ta[2]
            ac3dx = tc[0] - ta[0]
            ac3dy = tc[1] - ta[1]
            ac3dz = tc[2] - ta[2]
            bc3dx = tc[0] - tb[0]
            bc3dy = tc[1] - tb[1]
            bc3dz = tc[2] - tb[2]
            abtcx = tb[6]*self._texw - ta[6]*self._texw
            abtcy = tb[7]*self._texh - ta[7]*self._texh
            actcx = tc[6]*self._texw - ta[6]*self._texw
            actcy = tc[7]*self._texh - ta[7]*self._texh
            bctcx = tc[6]*self._texw - tb[6]*self._texw
            bctcy = tc[7]*self._texh - tb[7]*self._texh
            ab3dlen = math.sqrt(ab3dx*ab3dx + ab3dy*ab3dy + ab3dz*ab3dz)
            ac3dlen = math.sqrt(ac3dx*ac3dx + ac3dy*ac3dy + ac3dz*ac3dz)
            bc3dlen = math.sqrt(bc3dx*bc3dx + bc3dy*bc3dy + bc3dz*bc3dz)
            abtclen = math.sqrt(abtcx*abtcx + abtcy*abtcy)
            actclen = math.sqrt(actcx*actcx + actcy*actcy)
            bctclen = math.sqrt(bctcx*bctcx + bctcy*bctcy)
            t3dlen = (ab3dlen + ac3dlen + bc3dlen)/3.0
            ttclen = (abtclen + actclen + bctclen)/3.0
            if t3dlen > 0.0:
                ratio += ttclen/t3dlen
                ratiocnt += 1
        if ratiocnt > 0:
            imgscale = ratio/ratiocnt*1.2
        else:
            imgscale = 256
        dlg = QDialog(self)
        dlg.setWindowTitle('Triplanar Paste')
        layout = QFormLayout(dlg)
        smoothSlider = QSlider(Qt.Horizontal, dlg)
        smoothSlider.setMinimum(1000)
        smoothSlider.setMaximum(10000)
        smoothSlider.setValue(3000)
        layout.addRow('Hardness:', smoothSlider)
        scaleSpinBox = QDoubleSpinBox(dlg)
        scaleSpinBox.setMinimum(0.125)
        scaleSpinBox.setMaximum(4096)
        scaleSpinBox.setValue(imgscale)
        layout.addRow('Pixels per unit:', scaleSpinBox)
        moatSpinBox = QDoubleSpinBox(dlg)
        moatSpinBox.setValue(4)
        layout.addRow('Moat around geometry:', moatSpinBox)
        xBlendSpinBox = QDoubleSpinBox(dlg)
        xBlendSpinBox.setMinimum(0.0)
        xBlendSpinBox.setMaximum(1.0)
        xBlendSpinBox.setValue(1.0)
        xBlendSpinBox.setSingleStep(0.1)
        yBlendSpinBox = QDoubleSpinBox(dlg)
        yBlendSpinBox.setMinimum(0.0)
        yBlendSpinBox.setMaximum(1.0)
        yBlendSpinBox.setValue(1.0)
        yBlendSpinBox.setSingleStep(0.1)
        zBlendSpinBox = QDoubleSpinBox(dlg)
        zBlendSpinBox.setMinimum(0.0)
        zBlendSpinBox.setMaximum(1.0)
        zBlendSpinBox.setValue(1.0)
        zBlendSpinBox.setSingleStep(0.1)
        box = QHBoxLayout()
        box.addWidget(QLabel('X:'))
        box.addWidget(xBlendSpinBox)
        box.addWidget(QLabel('Y:'))
        box.addWidget(yBlendSpinBox)
        box.addWidget(QLabel('Z:'))
        box.addWidget(zBlendSpinBox)
        layout.addRow('Per-direction blend:', box)
        ignoreNegativeX = QCheckBox('Ignore negative X (for mirrored objects)')
        layout.addRow(ignoreNegativeX)
        box = QHBoxLayout()
        box.addStretch(1)
        btn = QPushButton('&OK', dlg)
        btn.setDefault(True)
        btn.clicked.connect(dlg.accept)
        box.addWidget(btn)
        btn = QPushButton('&Cancel', dlg)
        btn.clicked.connect(dlg.reject)
        box.addWidget(btn)
        layout.addRow(box)
        if dlg.exec() != QDialog.Accepted:
            return
        newrgba = bytearray(b'\x00'*(self._texw*self._texh*4))
        tgti = 0
        uvw = [0.0, 0.0, 0.0]
        gidx = 0
        y1 = 0
        blendpow = smoothSlider.value()/1000.0
        imgscale = scaleSpinBox.value()
        xblendfact = xBlendSpinBox.value()
        yblendfact = yBlendSpinBox.value()
        zblendfact = zBlendSpinBox.value()
        ignorenegx = ignoreNegativeX.isChecked()
        progress = QProgressDialog('Pasting, projecting and blending', 'Cancel', 0, self._xyzGridSize, self, )
        progress.show()
        cancelled = False
        for gt in range(0, self._xyzGridSize):
            progress.setValue(gt)
            QCoreApplication.instance().processEvents()
            if progress.wasCanceled():
                return
            x1 = 0
            for gs in range(0, self._xyzGridSize):
                c = self._xyzGrid[gidx]
                gidx += 1
                if c != None:
                    x2 = min(x1 + XYZ_GRID_SIZE, self._texw)
                    y2 = min(y1 + XYZ_GRID_SIZE, self._texh)
                    for y in range(y1, y2):
                        for x in range(x1, x2):
                            triFound = False
                            xx = x/self._texw
                            yy = y/self._texh
                            for t in c.tris:
                                ta = self._verts[t]
                                tb = self._verts[t + 1]
                                tc = self._verts[t + 2]
                                self.triSTBarycentric(ta, tb, tc, xx, yy, uvw)
                                if (uvw[0] >= 0.0) and (uvw[1] >= 0.0) and (uvw[2] >= 0.0):
                                    if ignorenegx and (ta[0]*uvw[0] + tb[0]*uvw[1] + tc[0]*uvw[2] < 0.0):
                                        continue
                                    triFound = True
                                    break
                            if not triFound:
                                for t in c.tris:
                                    ta = self._verts[t]
                                    tb = self._verts[t + 1]
                                    tc = self._verts[t + 2]
                                    self.triSTBarycentric(ta, tb, tc, xx, yy, uvw)
                                    if uvw[0] < 0.0:
                                        v1x = tc[6] - tb[6]
                                        v1y = tc[7] - tb[7]
                                        v2x = xx - tb[6]
                                        v2y = yy - tb[7]
                                        uvw[2] = min(1.0, max(0.0, (v2x*v1x + v2y*v1y)/(v1x*v1x + v1y*v1y)))
                                        uvw[1] = 1.0 - uvw[2]
                                        uvw[0] = 0.0
                                    elif uvw[1] < 0.0:
                                        v1x = ta[6] - tc[6]
                                        v1y = ta[7] - tc[7]
                                        v2x = xx - tc[6]
                                        v2y = yy - tc[7]
                                        uvw[0] = min(1.0, max(0.0, (v2x*v1x + v2y*v1y)/(v1x*v1x + v1y*v1y)))
                                        uvw[2] = 1.0 - uvw[0]
                                        uvw[1] = 0.0
                                    else:
                                        v1x = tb[6] - ta[6]
                                        v1y = tb[7] - ta[7]
                                        v2x = xx - ta[6]
                                        v2y = yy - ta[7]
                                        uvw[1] = min(1.0, max(0.0, (v2x*v1x + v2y*v1y)/(v1x*v1x + v1y*v1y)))
                                        uvw[0] = 1.0 - uvw[1]
                                        uvw[2] = 0.0
                                    if ignorenegx and (ta[0]*uvw[0] + tb[0]*uvw[1] + tc[0]*uvw[2] < 0.0):
                                        continue
                                    cls = (ta[6]*uvw[0] + tb[6]*uvw[1] + tc[6]*uvw[2])*self._texw - x
                                    clt = (ta[7]*uvw[0] + tb[7]*uvw[1] + tc[7]*uvw[2])*self._texw - y
                                    if cls*cls + clt*clt < 16:
                                        triFound = True
                                        break
                            if triFound:
                                px = ta[0]*uvw[0] + tb[0]*uvw[1] + tc[0]*uvw[2]
                                py = ta[1]*uvw[0] + tb[1]*uvw[1] + tc[1]*uvw[2]
                                pz = ta[2]*uvw[0] + tb[2]*uvw[1] + tc[2]*uvw[2]
                                onx = ta[3]*uvw[0] + tb[3]*uvw[1] + tc[3]*uvw[2]
                                ony = ta[4]*uvw[0] + tb[4]*uvw[1] + tc[4]*uvw[2]
                                onz = ta[5]*uvw[0] + tb[5]*uvw[1] + tc[5]*uvw[2]
                                nx = pow(abs(onx), blendpow)
                                ny = pow(abs(ony), blendpow)
                                nz = pow(abs(onz), blendpow)
                                nl = math.sqrt(nx*nx + ny*ny + nz*nz)
                                if nl < 0.0001:
                                    continue
                                nx /= nl
                                ny /= nl
                                nz /= nl

                                if onx < 0.0:
                                    rgba_x = img.pixelColor(int(pz*imgscale + 10000)%img.width(), int(-py*imgscale + 10000)%img.height())
                                else:
                                    rgba_x = img.pixelColor(int(-pz*imgscale + 10000)%img.width(), int(-py*imgscale + 10000)%img.height())
                                if ony < 0.0:
                                    rgba_y = img.pixelColor(int(px*imgscale + 10000)%img.width(), int(-pz*imgscale + 10000)%img.height())
                                else:
                                    rgba_y = img.pixelColor(int(px*imgscale + 10000)%img.width(), int(pz*imgscale + 10000)%img.height())
                                if onz < 0.0:
                                    rgba_z = img.pixelColor(int(-px*imgscale + 10000)%img.width(), int(-py*imgscale + 10000)%img.height())
                                else:
                                    rgba_z = img.pixelColor(int(px*imgscale + 10000)%img.width(), int(-py*imgscale + 10000)%img.height())

                                xblend = nx*nx*xblendfact
                                yblend = ny*ny*yblendfact
                                zblend = nz*nz*zblendfact

                                fr = rgba_x.redF()*xblend + rgba_y.redF()*yblend + rgba_z.redF()*zblend
                                fg = rgba_x.greenF()*xblend + rgba_y.greenF()*yblend + rgba_z.greenF()*zblend
                                fb = rgba_x.blueF()*xblend + rgba_y.blueF()*yblend + rgba_z.blueF()*zblend
                                fa = rgba_x.alphaF()*xblend + rgba_y.alphaF()*yblend + rgba_z.alphaF()*zblend

                                tgti = (y*self._texw + x)*4
                                newrgba[tgti] = int(max(0, min(255, fb*255.0)))
                                newrgba[tgti + 1] = int(max(0, min(255, fg*255.0)))
                                newrgba[tgti + 2] = int(max(0, min(255, fr*255.0)))
                                newrgba[tgti + 3] = int(max(0, min(255, fa*255.0)))
                x1 += XYZ_GRID_SIZE
            y1 += XYZ_GRID_SIZE
        node = doc.createNode('Triplanar paste', 'paintlayer')
        node.setOpacity(255)
        self.applySelectionAsAlphaMask(newrgba)
        node.setPixelData(QByteArray(newrgba), 0, 0, self._texw, self._texh)
        doc.rootNode().addChildNode(node, None)
        doc.refreshProjection()
        progress.deleteLater()

    def setTextureToCurrentDocument(self):
        self._doc = KI.activeDocument()
        self.repaint()

    def setTextureToAnyDocument(self):
        self._doc = KI.activeDocument()
        self.repaint()

    def snapshotTo50Percent(self):
        self._nextUpdateIsSnapshot = 2
        self.updateTexture()
        self.repaint()

    def snapshotTo25Percent(self):
        self._nextUpdateIsSnapshot = 4
        self.updateTexture()
        self.repaint()

    def snapshotTo125Percent(self):
        self._nextUpdateIsSnapshot = 8
        self.updateTexture()
        self.repaint()

    def initializeGL(self):
        context = self.context()
        context.aboutToBeDestroyed.connect(self.cleanupGL)
        profile = QOpenGLVersionProfile()
        profile.setVersion(2, 1)
        profile.setProfile(QSurfaceFormat.NoProfile)
        gl = self._gl = context.versionFunctions(profile)
        self._tex = 0
        self.updateTexture()
        self._tprog = QOpenGLShaderProgram();
        self._tprog.addShaderFromSourceCode(QOpenGLShader.Vertex,
# Texture coord projection vertex shader
"""
#version 120
varying vec2 vTexCoord;
void main()
{
    vTexCoord = gl_MultiTexCoord0.xy;
    gl_Position = ftransform();
}
""")
        self._tprog.addShaderFromSourceCode(QOpenGLShader.Fragment,
# Texture coord projection fragment shader
"""
#version 120
#extension GL_EXT_gpu_shader4 : enable
varying vec2 vTexCoord;
float to01(float v) { v = fract(v); return v < 0.0 ? 1.0 - v : v; }
void main()
{
    gl_FragColor = vec4(
        (int(to01(vTexCoord.x)*65535.0) & 0xFF)/255.0,
        (int(to01(vTexCoord.x)*65535.0) >> 8)/255.0,
        (int(to01(vTexCoord.y)*65535.0) & 0xFF)/255.0,
        (int(to01(vTexCoord.y)*65535.0) >> 8)/255.0);
}
""")
        self._tprog.link()
        self._updateTPData = True
        gl.glClearColor(self._viewBgColor.redF(), self._viewBgColor.greenF(), self._viewBgColor.blueF(), 1.0)

    def cleanupGL(self):
        gl = self._gl
        self.context().makeCurrent()
        gl.glDeleteTextures(1, [self._tex])
        gl.glDeleteLists(1, self._dl)

    def doTheRender(self):
        gl = self._gl
        gl.glClearColor(self._viewBgColor.redF(), self._viewBgColor.greenF(), self._viewBgColor.blueF(), 0.0)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPushMatrix()
        a = self.width()*1.0/self.height()
        if self._viewOrthographic:
            gl.glOrtho(-self._camd*0.5*a, self._camd*0.5*a, -self._camd*0.5, self._camd*0.5, 0.001, 1000.0)
        else:
            f = 1.0/math.tan(math.radians(45*0.5))
            gl.glLoadMatrixf([f/a, 0, 0, 0,
                0, f, 0, 0,
                0, 0, 1000.001/(0.001 - 1000.0), -1.0,
                0, 0, (2.0*1000.0*0.001)/(0.001 - 1000.0), 0])
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glPushMatrix()
        gl.glLoadIdentity()
        gl.glTranslatef(0, 0, -self._camd)
        gl.glRotatef(self._rotx, 1, 0, 0)
        gl.glRotatef(self._roty, 0, 1, 0)
        gl.glTranslatef(-self._tgtx, -self._tgty, -self._tgtz)
        gl.glColor3f(1, 1, 1)
        gl.glEnable(gl.GL_TEXTURE_2D)
        gl.glEnable(gl.GL_DEPTH_TEST)
        if self._viewFlat:
            gl.glDisable(gl.GL_LIGHTING)
        else:
            gl.glEnable(gl.GL_LIGHTING)
            gl.glEnable(gl.GL_LIGHT0)
        if self._viewBackfaces:
            gl.glDisable(gl.GL_CULL_FACE)
        else:
            gl.glEnable(gl.GL_CULL_FACE)
        gl.glBindTexture(gl.GL_TEXTURE_2D, self._tex)
        if self._viewFilter:
            gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR)
            gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR)
        else:
            gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR)
            gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_NEAREST)
        if self._dl == 0:
            self._dl = gl.glGenLists(1)
            gl.glNewList(self._dl, gl.GL_COMPILE)
            gl.glBegin(gl.GL_TRIANGLES)
            for v in self._verts:
                gl.glNormal3f(v[3], v[4], v[5])
                gl.glTexCoord2f(v[6], v[7])
                gl.glVertex3f(v[0], v[1], v[2])
            gl.glEnd()
            gl.glEndList()
        if self._updateTPData:
            gl.glClearColor(1.0, 1.0, 1.0, 1.0)
            gl.glClear(gl.GL_COLOR_BUFFER_BIT);
            self._updateTPData = False
            self._tprog.bind()
            gl.glCallList(self._dl)
            self._tprog.release()
            self._tp = gl.glReadPixels(0, 0, self.width(), self.height(), gl.GL_RGBA, gl.GL_FLOAT)
            gl.glClearColor(self._viewBgColor.redF(), self._viewBgColor.greenF(), self._viewBgColor.blueF(), 0.0)
            gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        gl.glCallList(self._dl)
        gl.glDisable(gl.GL_TEXTURE_2D)
        if self._viewWireframe:
            gl.glColor3f(1, 1, 1)
            gl.glDisable(gl.GL_LIGHTING)
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_LINE)
            gl.glEnable(gl.GL_POLYGON_OFFSET_LINE)
            gl.glPolygonOffset(self._camd*0.001, -1)
            gl.glCallList(self._dl)
            gl.glDisable(gl.GL_POLYGON_OFFSET_LINE)
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)
        gl.glPopMatrix()
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPopMatrix()
        gl.glMatrixMode(gl.GL_MODELVIEW)

    def paintGL(self):
        self.doTheRender()

    def resizeGL(self, width, height):
        gl = self._gl
        gl.glViewport(0, 0, width, height)


class View3DDocker(DockWidget):

    def __init__(self):
        super().__init__()
        self.setWindowTitle("View 3D")
        self._viewport = View3DViewport()

        layout = QVBoxLayout()
        layout.addWidget(self._viewport)

        buttons = QHBoxLayout()

        chk = QCheckBox("Live")
        chk.setCheckState(Qt.Checked)
        chk.stateChanged.connect(self.liveChanged)
        self._viewport._liveCheckbox = chk
        buttons.addWidget(chk)

        button = QPushButton("Update")
        button.clicked.connect(self._viewport.updateTexture)
        buttons.addWidget(button)

        buttons.addStretch()

        layout.addLayout(buttons)

        w = QWidget(self)
        w.setLayout(layout)
        self.setWidget(w)

    def liveChanged(self, state):
        self._viewport._periodicUpdates = state != Qt.Unchecked

    def canvasChanged(self, canvas):
        if self._viewport._gl:
            self._viewport.updateTexture()

Krita.instance().addDockWidgetFactory(DockWidgetFactory("view3d", DockWidgetFactoryBase.DockRight, View3DDocker))
